import { EditorState, basicSetup } from '@codemirror/basic-setup';
import { indentWithTab } from '@codemirror/commands';
import { EditorView, keymap } from '@codemirror/view';
import { reka } from '@rekajs/codemirror';
import { Parser } from '@rekajs/parser';
import { useCollector } from '@rekajs/react';
import * as t from '@rekajs/types';
import debounce from 'lodash/debounce';
import * as React from 'react';

/******************************************************************************
Copyright (c) Microsoft Corporation.

Permission to use, copy, modify, and/or distribute this software for any
purpose with or without fee is hereby granted.

THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT,
INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
PERFORMANCE OF THIS SOFTWARE.
***************************************************************************** */

var __assign = function() {
    __assign = Object.assign || function __assign(t) {
        for (var s, i = 1, n = arguments.length; i < n; i++) {
            s = arguments[i];
            for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p)) t[p] = s[p];
        }
        return t;
    };
    return __assign.apply(this, arguments);
};

function __read(o, n) {
    var m = typeof Symbol === "function" && o[Symbol.iterator];
    if (!m) return o;
    var i = m.call(o), r, ar = [], e;
    try {
        while ((n === void 0 || n-- > 0) && !(r = i.next()).done) ar.push(r.value);
    }
    catch (error) { e = { error: error }; }
    finally {
        try {
            if (r && !r.done && (m = i["return"])) m.call(i);
        }
        finally { if (e) throw e.error; }
    }
    return ar;
}

var _diffASTArrayTypes = function (program, newProgram, getTarget, isEqual) {
    var currentComponents = getTarget(program);
    var newComponents = getTarget(newProgram);
    var componentsToInsert = [];
    newComponents.forEach(function (newComponent, i) {
        var existingComponent = currentComponents.find(function (oldComponent) {
            return isEqual(oldComponent, newComponent);
        });
        if (!existingComponent) {
            componentsToInsert.push([newComponent, i]);
            return;
        }
        t.merge(existingComponent, newComponent);
    });
    componentsToInsert.forEach(function (_a, i) {
        var _b = __read(_a, 2), component = _b[0], index = _b[1];
        currentComponents.splice(index + i, 0, component);
    });
    currentComponents
        .filter(function (oldComponent) {
        return !newComponents.find(function (newComponent) {
            return isEqual(oldComponent, newComponent);
        });
    })
        .forEach(function (component) {
        currentComponents.splice(currentComponents.indexOf(component), 1);
    });
};
var diffAST = function (program, newProgram) {
    // Diff Globals
    _diffASTArrayTypes(program, newProgram, function (program) { return program.globals; }, function (a, b) { return a.name === b.name; });
    // Diff components
    _diffASTArrayTypes(program, newProgram, function (program) { return program.components; }, function (a, b) { return a.name === b.name; });
};
var CodeEditor = function (props) {
    var reka$1 = useCollector().reka;
    var domRef = React.useRef(null);
    var currentStateRef = React.useRef(t.Schema.fromJSON(reka$1.program));
    var currentCodeStringRef = React.useRef(Parser.stringify(reka$1.program));
    var isSynchingFromCodeMirror = React.useRef(false);
    var isSynchingFromExternal = React.useRef(false);
    var isTypingRef = React.useRef(false);
    // eslint-disable-next-line react-hooks/exhaustive-deps
    var syncCodeToState = React.useCallback(debounce(function (code) {
        var _a, _b;
        if (isSynchingFromExternal.current) {
            return;
        }
        isSynchingFromCodeMirror.current = true;
        try {
            var newAST_1 = Parser.parseProgram(code);
            reka$1.change(function () {
                diffAST(currentStateRef.current, newAST_1);
                diffAST(reka$1.program, currentStateRef.current);
            });
            (_a = props.onStatusChange) === null || _a === void 0 ? void 0 : _a.call(props, {
                type: 'success',
            });
        }
        catch (error) {
            (_b = props.onStatusChange) === null || _b === void 0 ? void 0 : _b.call(props, {
                type: 'error',
                error: error.message,
            });
        }
        isSynchingFromCodeMirror.current = false;
        isTypingRef.current = false;
    }, 1000), [reka$1]);
    var _a = __read(React.useState(null), 2), codemirrorView = _a[0], setCodemirrorView = _a[1];
    React.useLayoutEffect(function () {
        var dom = domRef.current;
        if (!dom) {
            return;
        }
        setCodemirrorView(new EditorView({
            state: EditorState.create({
                doc: currentCodeStringRef.current,
                extensions: [
                    basicSetup,
                    keymap.of([indentWithTab]),
                    reka(),
                    EditorView.theme({
                        '&': {
                            height: '100%',
                        },
                        '&.cm-focused': {
                            outline: 'none!important',
                        },
                        '.cm-scroller': {
                            'font-family': "'JetBrains Mono'",
                            fontSize: '0.875em',
                            lineHeight: '1.6em',
                            wordBreak: 'break-word',
                            '-webkit-font-smoothing': 'initial',
                        },
                    }),
                    EditorView.updateListener.of(function (view) {
                        var _a;
                        if (!view.docChanged || isSynchingFromExternal.current) {
                            return;
                        }
                        isTypingRef.current = true;
                        currentCodeStringRef.current = view.state.doc.toString();
                        (_a = props.onStatusChange) === null || _a === void 0 ? void 0 : _a.call(props, {
                            type: 'parsing',
                        });
                        syncCodeToState(currentCodeStringRef.current);
                    }),
                ],
            }),
            parent: dom,
        }));
    }, [syncCodeToState, props]);
    var onExternalChange = React.useCallback(function () {
        if (isSynchingFromCodeMirror.current || isTypingRef.current) {
            return;
        }
        if (!codemirrorView) {
            return;
        }
        if (isSynchingFromExternal.current === false) {
            isSynchingFromExternal.current = true;
            Promise.resolve().then(function () {
                var oldCode = currentCodeStringRef.current;
                var newCode = Parser.stringify(reka$1.program);
                if (newCode === oldCode) {
                    isSynchingFromExternal.current = false;
                    return;
                }
                currentStateRef.current = t.Schema.fromJSON(reka$1.program);
                var transaction = codemirrorView.state.update({
                    changes: {
                        from: 0,
                        to: codemirrorView.state.doc.length,
                        insert: newCode,
                    },
                });
                currentCodeStringRef.current = newCode;
                codemirrorView.dispatch(transaction);
                isSynchingFromExternal.current = false;
            });
        }
    }, [reka$1, codemirrorView]);
    // If the AST changes (ie: from undo/redo or from multiplayer),
    // Then, sync those changes to the CodeMirror editor
    React.useEffect(function () {
        var unsubscribe = reka$1.listenToChanges(function (payload) {
            if (payload.event !== 'change') {
                return;
            }
            onExternalChange();
        });
        return function () {
            unsubscribe();
        };
    }, [reka$1, codemirrorView, onExternalChange]);
    return React.createElement("div", __assign({}, props, { ref: domRef }));
};
CodeEditor.toString = function () { return '.reka-react-code-editor'; };

export { CodeEditor };
